from glue.config import settings, preference_panes
from qtpy import QtWidgets

from glue.config import viewer_tool
from glue.viewers.common.qt.tool import Tool

@viewer_tool
class GGUIAutoChop(Tool):

    icon = 'myicon.png'
    tool_id = 'custom_tool'
    action_text = 'Does cool stuff'
    tool_tip = 'Does cool stuff'
    shortcut = 'D'

    def __init__(self, viewer):
        super(MyCustomMode, self).__init__(viewer)

    def activate(self):
        pass

    def close(self):
        pass



"""
.. module:: make_param.py
    :synopsis: GUI used to create input param files needed to run GGUI.
.. moduleauthor:: Scott W. Fleming <fleming@stsci.edu>
"""
import csv
import os
import sys
import yaml
try:
    from PyQt5.QtCore import Qt
    from PyQt5.QtGui import QFont
    import PyQt5.QtWidgets as QtWidgets
    from PyQt5.QtWidgets import QApplication, QLineEdit, QPushButton, QFileDialog
    from PyQt5.QtWidgets import QTextEdit, QScrollArea
except ImportError:
    from PyQt4.QtCore import Qt
    from PyQt4.QtGui import QFont
    import PyQt4.QtWidgets as QtWidgets
    from PyQt4.QtWidgets import QApplication, QLineEdit, QPushButton, QFileDialog
    from PyQt4.QtWidgets import QTextEdit, QScrollArea

import pathlib
def validate_targlist_format(target_list: dict,  list_source: str) -> dict:
    """Verifies dictionary is in gGui YAML format
    Verifies a given dictionary is the established gGui format.
    Also verifies all specified files exist. Removes any target that does not have any valid files

    :returns: verified gGui target dictionary
    """
    empty_targets = []
    for target_name, target_data in target_list.items():
        valid_files = 0
        for data_type, band_data in target_data.items():
            if data_type == '_notes':
                continue
            for band, filepathString in band_data.items():
                if not pathlib.Path(filepathString).is_file():
                    if filepathString:
                        print(filepathString + " does not exist on disk. Ignoring...")
                else: valid_files += 1
        if not valid_files:
            empty_targets.append(target_name)

    for bad_target in empty_targets:
        print(str(bad_target) + " does not have any valid data. Ignoring target...")
        del target_list[bad_target]

    return target_list

def quoted_presenter(dumper, data):
    """
    This is a custom representer for string, so that YAML can be output
    with quotes to satisfy paths on Windows.
    """
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='"')

class MakeParamGUI(QtWidgets.QWidget):
    """
    This class builds a pyQt GUI for creating a GGUI YAML parameter file.
    """

    def __init__(self):
        yaml.add_representer(str, quoted_presenter)
        super().__init__()
        self.input_fname = ''
        self.output_fname = ''
        self.output_save_text = ''
        self.input_save_text = ''
        self.log_display = ''
        self.load_button = QPushButton()
        self.save_button = QPushButton()
        self.iscrollarea = QScrollArea()
        self.yaml_params = dict()
        self.init_ui()

    def getifile(self):
        """
        Retrieves selected file names and populates text box area with it.
        """
        self.input_fname, _filter = QFileDialog.getOpenFileName(
            self, 'Select Input File')
        self.input_save_text.setPlaceholderText(self.input_fname)

    def update_ofile(self):
        """
        Updates the output file name if text is entered or changed in the box.
        """
        self.output_fname = self.output_save_text.text().strip()
        # If the input_fname is set, then activate the save button.
        if self.input_fname:
            self.save_button.setStyleSheet(
                "QPushButton { background-color:"
                '"#7af442";border-color:"#45a018";}')
            self.save_button.setEnabled(True)
        # If the user has deleted all the text, disable the button.
        if not self.output_fname:
            self.save_button.setStyleSheet("QPushButton { background-color:"
                                           '"#FF4242";border-color:"#45a018";}')
            self.save_button.setEnabled(False)

    def make_output_row(self):
        """
        Create the row containing the output control.
        """
        # This creates the label itself.
        output_save_label = QtWidgets.QLabel('Output File Name: ')
        output_save_label.setAlignment(Qt.AlignLeft)
        output_save_label.setFont(QFont("Arial", 14, QFont.Bold))
        self.grid.addWidget(output_save_label, 0, 0, 1, 1)
        # This creates the current save file text area.
        self.output_save_text = QLineEdit()
        self.output_save_text.setReadOnly(False)
        self.output_save_text.setStyleSheet("border-style: solid; "
                                            "border-width: 1px; "
                                            "background: white; ")
        self.output_save_text.textChanged.connect(self.update_ofile)
        self.grid.addWidget(self.output_save_text, 0, 1, 1, 1)
        # This is the Save button for the output file.
        self.save_button = QPushButton("Save")
        self.save_button.setStyleSheet("QPushButton { background-color:"
                                       '"#FF4242";border-color:"#45a018";}')
        self.save_button.setEnabled(False)
        self.save_button.clicked.connect(self.save_yaml)
        self.grid.addWidget(self.save_button, 0, 2, 1, 1)

    def load_file(self):
        """
        Actions performed when pressing the Load button.
        Any row that begins with a pound sign charater is skipped.
        """
        if self.input_fname:
            with open(self.input_fname, 'r') as input_file:
                csv_reader = csv.reader(input_file, delimiter=',')
                input_fname_root = os.path.join(
                    os.path.dirname(self.input_fname), '')
                row_num = 0
                for row in csv_reader:
                    row_num = row_num + 1
                    if row[0].lstrip()[0] != '#':
                        if len(row) == 7:
                            # Then found expected number of entries.  Add this
                            # target to the list object to prepare for yaml
                            # output.
                            testfs = row[1:]

                            # Check for file existence.
                            for testf in testfs:
                                if not os.path.isfile(testf) and testf:
                                    self.log_display.append("File not found: "
                                                            + testf)
                            prikey = row[0].strip()
                            self.yaml_params[prikey] = {'lightcurve':{},
                                                        'coadd':{}, 'cube':{}}
                            if row[1].strip():
                                self.yaml_params[prikey]['lightcurve']['FUV'] = (
                                    os.path.abspath(input_fname_root +
                                                    row[1].strip()))
                            if row[2].strip():
                                self.yaml_params[prikey]['lightcurve']['NUV'] = (
                                    os.path.abspath(input_fname_root +
                                                    row[2].strip()))
                            if row[3].strip():
                                self.yaml_params[prikey]['coadd']['FUV'] = (
                                    os.path.abspath(input_fname_root +
                                                    row[3].strip()))
                            if row[4].strip():
                                self.yaml_params[prikey]['coadd']['NUV'] = (
                                    os.path.abspath(input_fname_root +
                                                    row[4].strip()))
                            if row[5].strip():
                                self.yaml_params[prikey]['cube']['FUV'] = (
                                    os.path.abspath(input_fname_root +
                                                    row[5].strip()))
                            if row[6].strip():
                                self.yaml_params[prikey]['cube']['NUV'] = (
                                    os.path.abspath(input_fname_root +
                                                    row[6].strip()))
                        else:
                            # Then need to print an error to the log area.
                            err_string = ("Row number {0:d}".format(row_num) +
                                          " does not have seven columns.")
                            self.log_display.append(err_string)
                # After parsing the CSV, enable the Save button.
                # Only enable if an output file has been decided.
                if self.output_fname:
                    self.save_button.setStyleSheet(
                        "QPushButton { background-color:"
                        '"#7af442";border-color:"#45a018";}')
                    self.save_button.setEnabled(True)

    def save_yaml(self):
        """
        Writes yaml parameters to file.
        """
        with open(self.output_fname, 'w') as outfile:
            yaml.dump(self.yaml_params, outfile, default_flow_style=False)

    def make_input_row(self):
        """
        Create the row containing the input control.
        """
        # This creates the label itself.
        input_save_label = QtWidgets.QLabel('Input File Name: ')
        input_save_label.setAlignment(Qt.AlignLeft)
        input_save_label.setFont(QFont("Arial", 14, QFont.Bold))
        self.grid.addWidget(input_save_label, 1, 0, 1, 1)
        # This creates the current input file name that was loaded.
        self.input_save_text = QLineEdit()
        self.input_save_text.setReadOnly(True)
        self.input_save_text.setStyleSheet("border-style: solid; "
                                           "border-width: 1px; "
                                           "background: white; ")
        self.iscrollarea = QScrollArea(self)
        self.iscrollarea.setWidget(self.input_save_text)
        self.iscrollarea.setWidgetResizable(True)
        self.grid.addWidget(self.iscrollarea, 1, 1, 1, 1)
        # This creates the file picker button to select the input file.
        self.load_button = QPushButton("Load")
        self.load_button.clicked.connect(self.getifile)
        self.load_button.setStyleSheet("QPushButton { background-color:"
                                       '"#7af442";border-color:"#45a018";}')
        self.load_button.clicked.connect(self.load_file)
        self.grid.addWidget(self.load_button, 1, 2, 1, 1)

    def make_errorlog_row(self):
        """
        Create the row containing the error log area.
        """
        self.log_display = QTextEdit()
        self.log_display.setReadOnly(True)
        self.log_display.setLineWrapMode(QTextEdit.NoWrap)
        self.log_display.setStyleSheet("border-style: solid; "
                                       "border-width: 1px; "
                                       "background: white; ")
        self.grid.addWidget(self.log_display, 2, 0, 1, 3)

    def init_ui(self):
        """
        Defines the GUI layout, text fields, buttons, etc.
        """
        # Create a grid layout and add all the widgets.
        self.grid = QtWidgets.QGridLayout()
        self.grid.setSpacing(10)

        # Create the row that contains the parameter file save options.
        self.make_output_row()

        # Create the row that contains the input upload controls.
        self.make_input_row()

        # Create the row that contains the error log area.
        self.make_errorlog_row()

        # Display the GUI.
        self.setLayout(self.grid)
        self.resize(800, 300)
        self.setWindowTitle("Create GGUI Input File")
        self.show()

if __name__ == "__main__":
    APP = QApplication(sys.argv)
    W = MakeParamGUI()
    sys.exit(APP.exec_())



"""
.. module:: autochop.py
    :synopsis: Determines time windows of GALEX observations based on time 
.. moduleauthor:: Duy Nguyen <dtn5ah@virginia.edu>
"""
import numpy
from glue.core import Data

def lightcurveChopList(parentData, axis, timeInterval):
    """
    Breaks observation data into observations separated by timeInterval. Returns list of times
    
    :param parentData: Glue (Pandas) Data Object containing CSV lightcurve data
    :type parentData: glue.core.data.Data

    :param axis: parameter to split across (usual = time)
    :type axis: string

    :param timeInterval: interval/amount to split parameter 'axis' across
    :type timeInterval: numpy.float64

    :returns: list -- List of autochop regions
    """
    import numpy
    # Calculate all time differences between points
    timeDifferences = numpy.diff(parentData[axis])
    # Find all indices of blank jumps greater than timeInterval
    obsWindows = []
    # Determine first observation window
    obsStart = parentData[axis, 0]
    for index, difference in enumerate(timeDifferences):
       # If the time difference is larger than specified time 
       if difference > timeInterval:
            # End time is the current index
            obsEnd = parentData['MeanTime', index]
            # Append this time window as tuple to master array
            obsWindows.append((index, obsStart, obsEnd))
            # Next window begins in next data point
            obsStart = parentData['MeanTime', index + 1]
    # Add last window to master array
    obsWindows.append((len(parentData[axis]), obsStart, parentData['MeanTime', -1]))
    return obsWindows

def lightcurveChopImport(glueApp, dataCollection, parentData, obsWindows):
    """
    Receives list of obs windows, breaks dataseries accordingly, imports data object to collection
    
    :param glueApp: Current instantiation of the Glue Application to spawn canvas into
    :type glueApp: glue.app.qt.application.GlueApplication

    :param dataCollection: Library of imported data objects to current Glue interface
    :type dataCollection: glue.core.data.DataCollection 

    :param parentData: Glue (Pandas) Data Object containing CSV lightcurve data
    :type parentData: glue.core.data.Data

    :param obsWindows: List of observation windows with indices upon which to chop
    :type obsWindows: list
    """
    from glue.core import Data
    # Set Initial Window start to index 0
    indxStart = 0
    # Get List of all extensions
    extensionList = parentData.component_ids()
    for window in obsWindows:
        # Grab ending index from list
        indxEnd = window[0]
        # Instantiate new Data Object container for the chop
        newChop = Data(label="AutoChop " + str(indxStart))
        for extension in extensionList:
            # Break every extension into segments defined by indxStart and indxEnd
            newChop[str(extension)] = parentData[extension][indxStart:indxEnd]
        # Import completed chop into dataCollection
        dataCollection.append(newChop)
        # Display Chop (for debugging purposes)
        #create_scatter_canvas(newChop,'MeanTime','Flux_BackgroundSubtracted',glueApp)
        # Move new index start to next index beyond current window
        indxStart = indxEnd + 1


# Before code restructuring, this was the only method here. Not sure what this is supposed to do in context of the other autochop methods...
def lightcurveChop(parentData, axis, timeInterval):
    
    # Calculate all time differences between points
    timeDifferences = numpy.diff(parentData[axis])
    # Find all indices of blank jumps greater than timeInterval
    obsWindows = []
    obsStart = parentData[axis, 0]
    for index, difference in enumerate(timeDifferences):
        if difference > timeInterval:
            obsEnd = parentData['MeanTime', index]
            obsWindows.append((obsStart,obsEnd))
            obsStart = parentData['MeanTime', index + 1]
    obsWindows.append((obsStart, parentData['MeanTime', -1]))

    return obsWindows
    """
    chopMeanTimes = []
    for chopIndex in chopIndices:
        chopMeanTimes.append(parentData['MeanTime',chopIndex])
    import ipdb; ipdb.set_trace()
    """

# Note for later. This is how you autochop :P
#from autochop import lightcurveChopList, lightcurveChopImport
#obsWindows = lightcurveChopList(lightcurveData, "MeanTime", 3600)
#lightcurveChopImport(glueApp, dataCollection, lightcurveData, obsWindows)


"""
.. module:: ggui
    :synopsis: Defines the gGui class and startup behavior
.. moduleauthor:: Duy Nguyen <dnguyen@nrao.edu>
"""

import argparse
import pathlib
import webbrowser
import yaml

from glue.core import DataCollection
from glue.app.qt.application import GlueApplication
from glue.config import menubar_plugin
from glue.utils import nonpartial
from PyQt5 import QtWidgets

from ggui import qtTabLayouts
from ggui.targetManager import TargetManager
from ggui.make_param import validate_targlist_format
from .version import __version__


class gGuiGlueApplication(GlueApplication):
    """Primary gGui Application Class
    Integrates gGui framework (target manager, custom tab generation, etc.) into Glue
    """

    def __init__(
        self,
        data_collection: DataCollection = DataCollection(),
        imported_target_catalogs: dict = None,
    ):
        """Initializes gGui
        If provided a dictionary of targets, in outlined gGui YAML structure,
        it will load those targets into the target manager

        :param data_collection: Glue data collection containing Glue data to plot
        :param imported_target_catalogs: Dict of targets and paths to associated gPhoton data products to load initially
        """
        super().__init__(data_collection)
        # Modify window title to specify gGui modified Glue environment
        self.setWindowTitle("gGui: gPhoton Graphical User Interface")
        # Add gGui YAML loader to "File" Menu
        self.menuBar().actions()[0].menu().addSeparator()
        self.menuBar().actions()[0].menu().addAction(
            "Load gGui Target Catalog", self.load_ggui_yaml
        )

        # Rename Glue "Help" to "Glue Help"
        self.menuBar().actions()[6].setText("&Glue Help")

        # Add "gGui Help" Menu
        menu_about_ggui = self.menuBar().addMenu("&gGui Help")
        # Add link to gGui ReadTheDocs
        ggui_rtd = QtWidgets.QAction("gGui &Online Documentation", menu_about_ggui)
        ggui_rtd.triggered.connect(
            nonpartial(webbrowser.open, "https://ggui.readthedocs.io/")
        )
        menu_about_ggui.addAction(ggui_rtd)
        # Add basic About information
        menu_about_ggui.addAction("About gGui", self.show_about_ggui)

        # Save a reference to the default tab
        # We won't need this, but can't delete it until we have multiple tabs
        default_tab = self.current_tab

        # Initialize blank overview tab
        def init_overview_tab(self):
            # Initialize blank overview_widget
            self.overview_widget = qtTabLayouts.ggui_overview_tab(session=self.session)
            self.tab_widget.addTab(
                self.overview_widget, "gGui Overview Tab: No Data Loaded"
            )
            # Set Overview Tab to focus
            self.tab_widget.setCurrentWidget(self.overview_widget)

        init_overview_tab(self)

        # Initialize empty Target Manager
        self.target_manager = TargetManager(self, self.primary_target_changed)
        self.addToolBarBreak()
        self.addToolBar(self.target_manager)

        if imported_target_catalogs:
            # Load supplied target catalog into target manager
            # NOTE: Upon first load, Target Manager will automatically update target manager's
            #       primary target with the first entry in this dict.
            self.load_targets(imported_target_catalogs)

        # Delete first default tab
        self.close_tab(self.get_tab_index(default_tab), False)

    def closeEvent(self, event):
        """Handles subwindows when gGui is closed"""

        # Notify Target Manager of closing
        self.target_manager.close()

    def primary_target_changed(self, _):
        """Updates tab data of new primary target
        Indended as signal callback for the target manager to notify gGui of primary target changes
        """
        # Update overview tab with new target's data
        self.overview_widget.load_data(
            self.session,
            self.target_manager.getPrimaryName(),
            self.target_manager.getPrimaryData(),
        )
        self.tab_widget.setTabText(
            self.get_tab_index(self.overview_widget),
            "Overview of " + str(self.target_manager.getPrimaryName()),
        )

    def load_targets(self, imported_target_catalogs: dict):
        """
        Imports gGui-compliant data (see yaml standard) into target manager

        :param imported_target_catalogs: Dict of gGui yamls and filenames imported from main
        """
        for filename in imported_target_catalogs:
            self.target_manager.loadTargetDict(
                filename,
                imported_target_catalogs[filename]
            )

    def load_ggui_yaml(self):
        """Prompts user with File Dialog for gGui YAML Target List
        Validates the YAML file and loads it into the Target Manager
        """
        for ggui_yaml_file in gGuiGlueApplication.prompt_user_for_file(
            "Select GGUI YAML Target List", "gGUI YAML (*.yaml; *.yml)"
        ):
            self.target_manager.loadTargetDict(
                ggui_yaml_file, 
                validate_targlist_format(
                    yaml.load(open(ggui_yaml_file, "r"), Loader=yaml.BaseLoader),
                    ggui_yaml_file,
                )
            )

    def show_about_ggui(self):
        """Displays about gGui Message Box"""
        QtWidgets.QMessageBox.about(
            self,
            "About gGui",
            "gPhoton Graphical User Interface (gGui)\n"
            "Developed by Duy Nguyen, Scott Fleming\nVersion: "
            + __version__
            + "\n\ngGui is provided under the AURA Software License. "
            "Please see the included license for details.",
        )

    @staticmethod
    def prompt_user_for_file(dialog_caption: str, dialog_name_filter: str) -> list:
        """
        Modular QtWidget File-Selection Dialog to prompt user for file import.
        Returns array of filenames selected

        :param dialog_caption: Caption to display along top of file dialog window
        :param dialog_name_filter: Filters file dialog to certain extension
        """
        # Set File Dialog Options
        dialog = QtWidgets.QFileDialog(caption=dialog_caption)
        dialog.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
        dialog.setNameFilter(dialog_name_filter)
        # Prompt User for file to import
        dialog.exec_()
        # Get array of File Names
        filenames = dialog.selectedFiles()
        return filenames


def main(user_arguments: list = None):
    """Entry point/helper function to start ggui

    :param user_arguments: list of arguments, should simulate command line args. Use ['-h'] or ['--help'] for help documentation
    """
    # Initialize argument parser with arguments
    parser = argparse.ArgumentParser(
        description="gPhoton Graphical User Interface. An analysis package for GALEX gPhoton "
                    "data products"
    )
    parser.add_argument(
        "--target_list",
        nargs="+",
        help="Specify a path to a YAML style list of astronomical targets and associated gPhoton "
             "data products",
    )
    parser.add_argument(
        "--yaml_select",
        action="store_true",
        help="Spawns a file select dialog to choose a YAML style list of astronomical targets and "
             "associated gPhoton data products",
    )
    if user_arguments:
        args = parser.parse_args(user_arguments)
    else:
        args = parser.parse_args()

    target_data_products = {}

    # If the user specified a gGui YAML file, load its targets
    if args.target_list:
        for ggui_yaml_file in args.target_list:
            target_list_path = str(pathlib.Path(ggui_yaml_file).absolute())
            target_data_products[target_list_path] = validate_targlist_format(
                yaml.load(open(target_list_path, "r"), Loader=yaml.BaseLoader),
                target_list_path,
            )
    # If the user requested a file-selector dialog to select a gGui YAML file, display it and load
    #   its contents
    if args.yaml_select:
        x = QtWidgets.QApplication([])
        for ggui_yaml_file in gGuiGlueApplication.prompt_user_for_file(
            "Select GGUI YAML Target List", "gGUI YAML (*.yaml; *.yml)"
        ):
            target_list_path = str(pathlib.Path(ggui_yaml_file).absolute())
            target_data_products[target_list_path] = validate_targlist_format(
                yaml.load(open(target_list_path, "r"), 
                Loader=yaml.BaseLoader),
                target_list_path,
            )
    # If no targets were recognized, notify the user
    if not target_data_products:
        print("No yaml received. Starting empty gGui session...")
    # Initialize gGui with user-supplied targets, if any
    ggui_app = gGuiGlueApplication(imported_target_catalogs=target_data_products)
    # Start gGui
    ggui_app.start()


if __name__ == "__main__":
    main()



"""
.. module:: qtTabLayouts
    :synopsis: Defines the gGui Target Manager to manage multiple targets
.. moduleauthor:: Duy Nguyen <dnguyen@nrao.edu>
"""

from collections import OrderedDict
from configparser import ConfigParser
import itertools
from typing import Callable
from copy import copy
import pathlib
import yaml

from PyQt5 import QtWidgets, QtGui, QtCore
from glue.app.qt.application import GlueApplication
from glue.core.link_helpers import LinkSame
from glue.core.data_factories import load_data

from pkg_resources import resource_filename


class TargetManager(QtWidgets.QToolBar):
    """
    Class that handles the loading of gPhoton data and management of multiple
    gGui targets
    """
    def __init__(self, glue_parent: GlueApplication, target_change_callback: Callable[[str], None] = None):
        """Initializes gGui Target Manager
        If provided a dictionary of targets, in outlined gGui YAML structure, it will load those targets into the target manager

        :param glue_parent: The Glue instance that instantiated this object. Target Manager cannot be run outside a Glue context
        :param target_change_callback: Callback function to notify upon primary target change
        """
        super().__init__()
        self._glue_parent = glue_parent
        self._target_catalog = OrderedDict()
        self._primary_data = {}
        self._target_change_callbacks = []
        self._target_notes = None
        self._note_display_widget = target_note_display(self)

        # Initialize GUI Elements
        self.addWidget(QtWidgets.QLabel("gGui Target Manager: "))
        # Add Back Button
        self.addAction(QtGui.QIcon(resource_filename('ggui.icons', 'ArrowBack_transparent.png')), "Previous Target", self.previous_target)
        # Add Combo Box
        self.QComboBox = QtWidgets.QComboBox(self)
        self.QComboBox.currentIndexChanged.connect(self.setPrimaryTarget)
        self.addWidget(self.QComboBox)
        # Add Forward Button
        self.addAction(QtGui.QIcon(resource_filename('ggui.icons', 'ArrowForward_transparent.png')), "Next Target", self.next_target)
        # Add Info button
        self.addAction(QtGui.QIcon(resource_filename('ggui.icons', 'Information.svg')), "Target Information", self.show_targ_info)
        # Add Notes Button
        self.addAction(QtGui.QIcon(resource_filename('ggui.icons', 'Notepad.png')), "Target Notes", self._note_display_widget.show)

        # If the initializer wants to know about target changes, register its provided callback
        if target_change_callback:
            self.register_target_change_callback(target_change_callback)

    def close(self):
        """Handles graceful exit housekeeping"""

        # Close note display widget if open
        self._note_display_widget.close()

    def register_target_change_callback(self, callback):
        """Registers a callback function to call when primary target changes

        :param callback: Callback function
        """
        self._target_change_callbacks.append(callback)

    def loadTargetDict(self, target_catalog: str, target_files: dict):
        """Loads a single dictionary of targets and associated data product paths into internal cache

        :param target_files: gGui compliant yaml dictionary of targets and paths to associated gPhoton data products
        :param target_catalog: Name/identifier of this dictionary of targets. Can be used to return data
        """
        # Verify the catalog exists
        target_catalog = str(pathlib.Path(target_catalog).resolve())
        # Check if we've already loaded this file. Otherwise throw an error
        if target_catalog in self._target_catalog:
            raise ValueError("Duplicate gGui catalog. Catalog already imported into gGui: " + target_catalog)
        # Add catalog targets to internal cache
        self._target_catalog[target_catalog] = OrderedDict(target_files)
        # Add new items to GUI
        for item in target_files.keys():
            self.QComboBox.addItem(item, {'target_catalog': target_catalog})

    def setPrimaryTarget(self, targIndex: int):
        """Changes primary target to target specified
        Unloads existing primary target's data (internal cache and parent Glue session),
        loads the new primary target's data, links their corresponding attributes together,
        and notifies all stakeholders of the new changed primary target

        :param targIndex: Index of desired new primary target
        """
        targName = self.QComboBox.currentText()
        targ_catalog = self.QComboBox.currentData()['target_catalog']
        # If requested target is not in current cache, throw exception
        if targName not in self.getTargetNames():
            raise KeyError("Target Manager does not recognize requested target: " + str(targName))

        # If we have data loaded, remove it from the Glue Data
        if self._primary_data:
            def unload_primary_data():
                for band_data_set in list(self._primary_data.values()):
                    for band_data in band_data_set.values():
                        if isinstance(band_data, list):
                            for data in band_data:
                                self._glue_parent.data_collection.remove(data)   
                        else:
                            self._glue_parent.data_collection.remove(band_data)                
            unload_primary_data()

        # Save target notes
        self._note_display_widget.save_notes()

        # Clear internal target cache
        self._primary_data.clear()
        self._target_notes = None

        target_files = copy(self.getTargetFiles(targ_catalog, targName))
        self._target_notes = target_files.pop('_notes', None)

        # For each gGui Data Type...
        for data_product_type in target_files:
            # Initialize dictionary for this data product
            self._primary_data[data_product_type] = {}

            # Retrieve the x and y attributes for this data product from the conf file
            config = ConfigParser()
            config.read(resource_filename('ggui', 'ggui.conf'))                
            x_att = config.get('Mandatory Fields', data_product_type + "_x", fallback='')
            y_att = config.get('Mandatory Fields', data_product_type + "_y", fallback='')

            # Load every band's data into internal cache
            for band, band_file in target_files[data_product_type].items():
                if band_file:
                    self._primary_data[data_product_type][band] = load_data(band_file)

                    # If x_att, y_att provided in conf, test they exist
                    try:
                        if x_att:
                            self._primary_data[data_product_type][band].id[x_att]
                        if y_att:
                            self._primary_data[data_product_type][band].id[y_att]
                    # KeyError means specified attribute doesn't exist in this data. Warn user, and unset attributes, but continue
                    except KeyError as e:
                        parsed_error = e.args[0].split(':')
                        if parsed_error[0]== 'ComponentID not found or not unique':
                            print("WARNING: '" + parsed_error[1].strip() + "' field specified in ggui.conf missing from " + targName + " " + data_product_type + " " + band + ": " + band_file)
                            x_att = ''
                            y_att = ''
                        else:
                            raise
                    # If AttributeError, check if "data" is actually a list of data (multiple data sets per file). Breaks 1-1 correspondence gGui assumes. Warn user plotting and gluing will fail. Skip this data, but import it regardless
                    except AttributeError:
                        if isinstance(self._primary_data[data_product_type][band], list):
                            print("WARNING: " + str(len(self._primary_data[data_product_type][band])) + " datasets imported from " + targName + " " + data_product_type + " band " + band + ". gGui shall import this data, but will be unable to perform automatic actions on it (i.e. gluing, displaying overview, etc.)")
                        else:
                            raise
                        
                    # Register this data product with Glue's Data Collection
                    self._glue_parent.data_collection.append(self._primary_data[data_product_type][band])
            # If we have multiple bands, glue them together
            try: 
                if len(self._primary_data[data_product_type].keys()) > 1:
                    attributes_to_glue = {'lightcurve': ['t_mean', 'flux_bgsub'], 
                                        'coadd': ['Right Ascension', 'Declination'], 
                                        'cube': ['Right Ascension', 'Declination', 'World 0']}
                    #for glue_attribute in attributes_to_glue[data_product_type]:
                    for glue_attribute in list(filter(lambda x: x is not '', [x_att, y_att] + config.get('Additional Fields To Glue', data_product_type, fallback='').split(','))):
                        #Can only link two fields at a time. Need to go through all combinations
                        from itertools import permutations
                        for linking_pair in (set(frozenset(t) for t in permutations(self._primary_data[data_product_type].values(),2))):
                            accessor = tuple(linking_pair)
                            self._glue_parent.data_collection.add_link(LinkSame(accessor[0].id[glue_attribute],accessor[1].id[glue_attribute]))
            except TypeError as e:
                print("Unable to glue " + str(targName) + " " + str(data_product_type) + ": " + str(e))

        # Notify all stakeholders of target change
        for callback in self._target_change_callbacks:
            callback(self.getPrimaryName())

    def setPrimaryNotes(self, new_notes: str):
        """"
        Updates internal cache of target's notes to given string

        :param new_notes: New notes for the primary target
        """
        self._target_notes = new_notes
        self.getTargetFiles(self.getPrimaryTargetCatalog(), self.getPrimaryName())['_notes'] = new_notes

    def getTargetNames(self) -> list:
        """Returns the names of all registered targets, in their registered order

        :returns: list of all cached targets' names
        """
        # Devnote 8: List Comprehension Alternative
        return list(itertools.chain.from_iterable(self._target_catalog.values()))

    def getTargetFiles(self, target_catalog: str, target_name: str) -> dict:
        """Returns the files and metadata of a specified target (Unloaded data, as per lazy evaluation principle)
        
        :param target_catalog: gGui catalog file this target originated from
        :param target_name: Name of the target whose files to lookup
        :returns: Unloaded metadata and filepaths of the corresponding target's data
        """
        try:
            return self._target_catalog[target_catalog][target_name]
        except KeyError:
            raise KeyError("'" + str(target_name) + "' not found in cache")

    def getTargetNotes(self, target_catalog: str, target_name: str) -> str:
        """Returns the notes specified target, or blank string if no notes registered
        
        :param target_catalog: gGui catalog this target originated from
        :param target_name: Name of the target whose notes to lookup
        :returns: Notes of the specified target, or blank string if no notes
        """
        try:
            return self.getTargetFiles(target_catalog, target_name)['_notes']
        except KeyError:
            return ""

    def getPrimaryData(self) -> dict:
        """Returns currently loaded primary target's data

        :returns: dictionary of the current primary target's data
        """
        return self._primary_data

    def getPrimaryName(self) -> dict:
        """Returns the current primary target's name

        :returns: current primary target's name as string
        """
        return self.QComboBox.currentText()

    def getPrimaryTargetCatalog(self) -> str:
        """Returns the current primary target's parent gGui Target Catalog path
        If no target is selected, returns a blank string

        :returns: primary target's gGui Target Catalog path as string
        """
        try:
            return self.QComboBox.currentData()['target_catalog']
        except TypeError:
            return ""

    def getPrimaryNotes(self) -> str:
        """Returns any notes associated with the current target. Returns empty string if no notes found.

        :returns: notes registered with the current target
        """
        return self._target_notes

    def next_target(self):
        """Advances to next primary target"""
        # Determine the index we're switching to...
        current_target_index = self.QComboBox.currentIndex()
        next_target_index = current_target_index + 1
        # And wrap around to the front if we're currently on the last target
        if next_target_index > self.QComboBox.count() - 1:
            next_target_index = 0
        # Command Target Manager to switch primary targets
        self.QComboBox.setCurrentText(self.QComboBox.itemText(next_target_index)) # QComboBox signal will initiate primary target switching

    def previous_target(self):
        """Advances to previous primary target"""
        # Determine the index we're switching to...
        current_target_index = self.QComboBox.currentIndex()
        next_target_index = current_target_index - 1
        # And wrap around to the back if we're currently on the first target
        if next_target_index < 0:
            next_target_index = self.QComboBox.count() - 1
        # Command Target Manager to switch primary targets
        self.QComboBox.setCurrentText(self.QComboBox.itemText(next_target_index)) # QComboBox signal will initiate primary target switching

    def show_targ_info(self):
        """Displays name and target catalog for the primary target"""
        QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Information, 
            "About Target", 
            "Target name: " + str(self.getPrimaryName()) + 
            "\ngGui Target Catalog: " + str(self.getPrimaryTargetCatalog()), 
            QtWidgets.QMessageBox.Ok
        ).exec()
    
    def flushSourceFile(self, source_filename: str):
        """Force saves (flushes) the given source file
        Intended to be used for saving notes

        :param source_filename: Filename of source file to be flushed
        """
        with open(source_filename, "w") as source_file:
            source_file.write(yaml.dump(dict(self._target_catalog[source_filename])))

class target_note_display(QtWidgets.QGroupBox):
    """Subwidget to display notes of current target"""

    def __init__(self, parent):
        """
        Initializes note display widget

        :param parent: The parent that spawned this note widget (usually the target manager)
        """
        super().__init__()
        self.setWindowTitle("gGui Notepad")
        self._target_manager = parent
        self._target_manager.register_target_change_callback(self.primary_target_changed)
        # Initialize Widgets
        # Main Text Field
        self._text_field = QtWidgets.QTextEdit()
        self._text_field.document().modificationChanged.connect(self.modificationChanged)
        # Save Notes Button
        self._save_button = QtWidgets.QPushButton("Save Notes")
        self.setTitle("Notes: Unmodified")
        self._save_button.setEnabled(False)
        self._save_button.clicked.connect(lambda: self.save_notes(True))
        # Discard Changes Button
        self._discard_button = QtWidgets.QPushButton("Discard Changes")
        self._discard_button.clicked.connect(self.discard_note_changes)
        self._discard_button.setEnabled(False)
        # Declare Layout
        self._layout = QtWidgets.QGridLayout()
        self.setLayout(self._layout)
        # Organize Widgets in Layout
        self._layout.addWidget(self._text_field, 0, 0, 1, 2)
        self._layout.addWidget(self._save_button, 1, 0)
        self._layout.addWidget(self._discard_button, 1, 1)

    def closeEvent(self, _):
        """When close is detected, prompts user to save notes if text has been modified"""
        self.save_notes()

    def primary_target_changed(self, new_target: str):
        """
        When primary target has changed, retrieves notes for the new target

        :param new_target: Name of the new primary target
        """
        # Get the new notes and update our text field
        self._text_field.setText(self._target_manager.getPrimaryNotes())
        # Set text field to unmodified to recalibrate autosave detection
        self._text_field.document().setModified(False)

    def save_notes(self, force_save: bool = False):
        """
        Checks if notes need to be saved. If so, saves the notes and flushes to disk

        :param force_save: If True, skips text modification checks and forces a save to disk
        """
        # Check for abort-save conditions
        if not force_save:
            # Check if text has been modified
            unsaved_text = self._text_field.document().isModified()
            # If text has been modified, ask user if they want to save. Otherwise, fallthrough to save
            if unsaved_text:
                if QtWidgets.QMessageBox.Cancel == QtWidgets.QMessageBox.question(self, "Close Confirmation", "Do you want to save changes to your notes?", QtWidgets.QMessageBox.Save | QtWidgets.QMessageBox.Cancel):
                    return
            # If text hasn't been modified, exit
            else:
                return
        # If no abort-save conditions caught, save to disk
        self._target_manager.setPrimaryNotes(self._text_field.toPlainText())
        try:
            self._target_manager.flushSourceFile(self._target_manager.getPrimaryTargetCatalog())
            # Set text field to unmodified to recalibrate autosave detection
            self._text_field.document().setModified(False)
        except IOError:
            print("Error saving notes! Your notes have NOT been saved!")

    def discard_note_changes(self):
        """Discards any changes to notes and reverts to last saved notes"""

        # Safety Prompt
        if QtWidgets.QMessageBox.Cancel == QtWidgets.QMessageBox.question(self, "Discard Confirmation", "Are you sure you want to permanently discard changes to your notes?", QtWidgets.QMessageBox.Discard | QtWidgets.QMessageBox.Cancel):
            return
        # Revert Notes
        self._text_field.setText(self._target_manager.getPrimaryNotes())
        # Set text field to unmodified to recalibrate autosave detection
        self._text_field.document().setModified(False)
        self.setTitle("Notes: Changes Discarded")
        self.setStyleSheet('QGroupBox:title {color: rgb(0, 0, 175);}')

    def modificationChanged(self, changed: bool):
        self._save_button.setEnabled(changed)
        self._discard_button.setEnabled(changed)
        if changed:
            self.setTitle("Notes: Modified")
            self.setStyleSheet('QGroupBox:title {color: rgb(255, 0, 0);}')
        else:
            self.setTitle("Notes: Saved")
            self.setStyleSheet('QGroupBox:title {color: rgb(0, 150, 0);}')
